#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "psg.h"

#define MODE_SQUARE 0
#define MODE_PSGSQR 1
#define MODE_SINUS 2

double psg_one_sample(double in, int mode)
{
  double out,tmp;

  tmp = fmod(in, M_PI*2);

  switch(mode) {
  case MODE_SQUARE:
    if(tmp < M_PI)
      return -1;
    else
      return 1;
  case MODE_SINUS:
    return sin(tmp);
  case MODE_PSGSQR:
  default:
    if(tmp < M_PI)
      out = 2 / tmp;
    else
      out = -(2 / (tmp - M_PI));
    
    if(out > 1)
      out = 1;
    
    if(out < -1)
      out = -1;
    
    return out;
  }
}

INLINE int psgfreq_to_freq(int psgfreq)
{
  if(psgfreq)
    {
      int denominator = SAMPLE_FREQ * psgfreq;
      return (DETAIL *(PSG_FREQUENCY/16.0) + (denominator >> 1)) / denominator;
    }
  else
    return 0;
}

VBL_Frame *psg_data_to_frame(YM_Header *head, int frame_number, VBL_Frame *old)
{
  unsigned char *tmp, tmp2[16];
  int prediv[8] = {-1, 4, 10, 16, 50, 64, 100, 200 };
  
  VBL_Frame *new;
  int i, tmpval;

  new = (VBL_Frame *) malloc(sizeof(VBL_Frame));

  tmp = head->data + frame_number;
  /* tmp = tmp2; */

  for(i=0;i<head->regs;i++)
    tmp2[i] = tmp[i*head->length];

  for(i=0;i<MAX_CHANNELS;i++) {
    new->channel[i].psg_freq = tmp2[2*i] + ((tmp2[2*i+1]&0xf) << 8);
    new->channel[i].volume = 15-(((tmp2[8+i]&0x1f)==16)?15:(tmp2[8+i]&0x1f));
    new->channel[i].frequency = psgfreq_to_freq(new->channel[i].psg_freq);
    new->channel[i].period_count = (old->channel[i].period_count);
    new->channel[i].tone_enable = !(tmp2[7] & (1<<i));
    new->channel[i].noise_enable = !(tmp2[7] & (8<<i));
    new->channel[i].env_enable = ((tmp2[8+i]&0x1f) == 16);
    new->channel[i].timer_enable = 0;
  }

  if(head->regs > 14) {
    tmpval = (tmp2[1] & 0x30) >> 4;
    
    if(tmpval) {
      new->channel[tmpval-1].timer_enable = 1;

      tmpval = (tmp2[6] & 0xe0) >> 5; 
      new->timer_freq = (MFP_FREQUENCY/prediv[tmpval])/tmp2[14];
      new->timer_duration = (SAMPLE_FREQ/new->timer_freq);
      if(tmp2[1] & 0x40)
	new->timer_count = 0;
      else
	new->timer_count = old->timer_count;
      
      new->timer_vmax = 15; /* (tmp2[5] & 0xf0) >> 4; */
    }
  }

  new->env_psg_freq = tmp2[11] + (tmp2[12] << 8);
  if(new->env_psg_freq)
    {
      new->env_frequency = DIV (PSG_FREQUENCY, 256 * new->env_psg_freq);
      /* 
	 new->env_frequency =
	 (((double)PSG_FREQUENCY/(double)256)/(double)new->env_psg_freq);
	 */
    }
  else
    new->env_frequency = 0;

  if(tmp2[13] == 0xff) {
    new->env_mode = old->env_mode;
    fflush(stderr);
    new->env_period_count = old->env_period_count;
    new->env_period_hold = old->env_period_hold;
  } else {
    new->env_mode = tmp2[13];
    fflush(stderr);
    new->env_period_count = 0;
    new->env_period_hold = 0;
  }
  free(old);

  return new;
}

VBL_Frame *psg_empty_to_frame(void)
{
  VBL_Frame *new;
  int i;

  new = (VBL_Frame *) malloc(sizeof(VBL_Frame));
  new->env_period_count = 0;
  new->env_period_hold = 0;
  new->env_mode = 0;
  for(i=0;i<MAX_CHANNELS;i++) {
    new->channel[i].period_count = 0;
  }
  return new;
}

void psg_build_tables(YM_Header *head)
{
  int i;

  for(i=0;i<DETAIL;i++) {
    head->tone[i][MODE_SQUARE] = 
      (int)(32760*psg_one_sample(i*M_PI*2/DETAIL, MODE_SQUARE)/3 + 0.5);
    head->tone[i][MODE_PSGSQR] = 
      (int)(32760*psg_one_sample(i*M_PI*2/DETAIL, MODE_PSGSQR)/3 + 0.5);
    head->tone[i][MODE_SINUS] = 
      (int)(32760*psg_one_sample(i*M_PI*2/DETAIL, MODE_SINUS)/3 + 0.5);
    head->noise[i] = ((rand() & 65535) - 32768)/3;
  }
}

YM_Header *psg_build_header(unsigned char *data, int size, PSG_Settings *set)
{
  int type;
  YM_Header *head;
  
  head = (YM_Header *) malloc(sizeof(YM_Header));
  
  head->id = SWAPLONG(*((long *) data));
  type = ((head->id & 0xff00) >> 8) - '0';

  switch(type)
    {
    case 2:
    case 3:
      head->length = (size-4)/14;
      head->data = data + 4;
      head->regs = 14;
      break;
    case 4:
    case 5:
    case 6:
      head->length = SWAPLONG(*((long *) (data + 12)));
      head->mask = SWAPLONG(*((long *) (data + 16)));
      head->no_digi = SWAPSHORT(*((short *) (data + 20)));
      if(head->no_digi) {
	fprintf(stderr, "Digidrum is unsupported.\n");
	return 0;
      }
      head->regs = 16;
      head->name = (char *)data + 34;
      head->name2 = head->name + strlen(head->name) + 1;
      head->name3 = head->name2 + strlen(head->name2) + 1;
      head->data = (unsigned char *)head->name3 + strlen(head->name3) + 1;
      if(set->verbose) {
	fprintf(stderr, "1: %s\n", head->name);
	fprintf(stderr, "2: %s\n", head->name2);
	fprintf(stderr, "3: %s\n", head->name3);
      }
      break;
    default:
      return 0;
    }
  psg_build_tables(head);
  return head;
}

void psg_play_frame(YM_Header *head, VBL_Frame *frame, PSG_Settings *set,
		    int currframe, int echoframe)
{
  static signed short buffer[ECHO_DISTANCE+1][SAMPLE_BUFFER/VBL];
  int i,c,tmp,count,left,right,timer_vol;
  int sample[MAX_CHANNELS];

  for(i=0;i<(SAMPLE_FREQ/VBL);i++) {
    for(c=0;c<MAX_CHANNELS;c++) {
      tmp = 0;
      count = 0;
      if(frame->channel[c].timer_enable) {
	frame->timer_count = 
	  (frame->timer_count + 1)%(frame->timer_duration*2);
	if(frame->timer_count > frame->timer_duration) {
	  timer_vol = 0;
	} else {
	  timer_vol = frame->timer_vmax;
	}
      } else {
	timer_vol = frame->channel[c].volume;
      }
      if(frame->channel[c].tone_enable) {
	tmp = head->tone[frame->channel[c].period_count&DETAIL_LOOP]
	  [MODE_SQUARE] >> timer_vol;
	count++;
	frame->channel[c].period_count += frame->channel[c].frequency;
      }
      if(frame->channel[c].noise_enable) {
	tmp += head->noise[i&DETAIL_LOOP] >> timer_vol;
	count++;
      }
      if(frame->channel[c].env_enable) {
#ifdef FIXED_MATH
	const int factor = (int)(256 * M_PI*2/DETAIL + 0.5);
	tmp *= env_wave(frame->env_period_count*factor, frame->env_mode);
	tmp = FIXED_TO_INT (tmp);
#else
	tmp *= env_wave(frame->env_period_count*M_PI*2/DETAIL, frame->env_mode);
#endif
      }
#if 1
      if(count)
	sample[c] = tmp/count;
#else
      if (tmp > 32767/3)
	tmp = 32767/3;
      else if (tmp < -32768/3)
	tmp = -32768/3;
      sample[c] = tmp;
#endif
    }
    if(frame->env_frequency)
      frame->env_period_count =
	FIXED_TO_INT (FIXED (frame->env_period_count) +
		      2 * DIV (DETAIL, SAMPLE_FREQ * frame->env_frequency));
    /* frame->env_period_count += DETAIL/(SAMPLE_FREQ/frame->env_frequency)*2; */

    if(set->stereo) {
      left = 2*sample[0]+sample[1];
      right = sample[1]+2*sample[2];
    } else {
      left = right = sample[0]+sample[1]+sample[2];
    }

    if(set->echo) {
      if(set->stereo) {
	buffer[currframe][i*2] = (10 * left + ECHO_LEVEL * buffer[echoframe][i*2+1])/(ECHO_LEVEL+10);
	buffer[currframe][i*2+1] = (10 * right + ECHO_LEVEL * buffer[echoframe][i*2])/(ECHO_LEVEL+10);
      } else {
	buffer[currframe][i*2] = (10 * left + ECHO_LEVEL * buffer[echoframe][i*2])/(ECHO_LEVEL+10);
	buffer[currframe][i*2+1] = (10 * right + ECHO_LEVEL * buffer[echoframe][i*2+1])/(ECHO_LEVEL+10);
      }
    } else {
      buffer[currframe][i*2] = (signed short)left;
      buffer[currframe][i*2+1] = (signed short)right;
    }
  }
  play_sample(buffer[currframe], SAMPLE_BUFFER/VBL);
}

void psg_play(YM_Header *head, PSG_Settings *set)
{
  VBL_Frame *frame, *old;
  int i, sec=0, oldsec=0;

  old = psg_empty_to_frame();

  for(i=0;i<head->length;i++) {
    frame = psg_data_to_frame(head, i, old);

    if(set->echo)
      psg_play_frame(head, frame, set, i%(ECHO_DISTANCE+1),(i+ECHO_DISTANCE+2)%(ECHO_DISTANCE+1));
    else
      psg_play_frame(head, frame, set, 0, 0);
      
    old = frame;
    oldsec = sec;
    sec = 1+((head->length-i)/VBL);
    if((sec != oldsec) && set->verbose) {
      fprintf(stderr, "\rTime left: %02d:%02d", sec/60,sec%60);
      fflush(stderr);
    }
  }
  if(set->verbose && !set->quiet)
    fprintf(stderr, "\rTime left: %02d:%02d\n", 0, 0);
}
